"""Main Calcuation Module for Energy Sharing."""

import pandas as pd
from .models import CalculationMethod
from .const import GROSS_INJECTION, GROSS_OFFTAKE, KEY, NET_INJECTION, NET_OFFTAKE, SHARED_ENERGY
from .data_formatting import create_multi_index_output_frame, result_to_input_for_reiteration


def _calculate(df: pd.DataFrame, method: CalculationMethod) -> pd.DataFrame:
    """Calculate the energy sharing for the given input data. This function is not iterative."""
    # Step 1: Calculate the maximum available gross injection that can be shared
    # A participant cannot share their injection with themselves

    # Take the injection of P1, and divide it per participant as per their key

    injections_to_share = []
    rest = {}

    for participant in df[GROSS_INJECTION].columns:
        injection_to_share = df[GROSS_INJECTION][participant].copy()

        key = df[KEY].copy()
        if method == CalculationMethod.RELATIVE or method == CalculationMethod.OPTIMAL:
            # Set the key of the current participant to 0
            # Re-normalize the keys for the other participants
            if participant in df[KEY].columns:
                key.loc[:, participant] = 0
                key = key.div(key.sum(axis=1), axis=0)

        # Multiply injection_to_share with the key of each participant
        shared_by_participant = (injection_to_share * key.T).T
        shared_by_participant.fillna(0, inplace=True)

        # Set the value for the current participant to 0
        if participant in shared_by_participant.columns:
            shared_by_participant.loc[:, participant] = 0

        # Put the not shared injection in the rest
        rest[participant] = injection_to_share - shared_by_participant.sum(axis=1)

        injections_to_share.append(shared_by_participant)

    # Sum the injections to share
    max_allocated_injection = sum(injections_to_share)

    # Concat the rest
    injection_that_cannot_be_shared = pd.concat(rest, axis=1)

    # Step 2: Calculate the Net Offtake, by assigning the injections to each participant
    # But, a participant cannot receive more than their offtake

    net_offtake = df[GROSS_OFFTAKE] - max_allocated_injection

    # Sum all negative values into a column "Not Shared"
    not_shared_after_assignment = net_offtake.clip(upper=0).sum(axis=1).abs()

    # Clip the values to 0
    net_offtake = net_offtake.clip(lower=0)

    # Calculate the amount of actual shared energy
    # This is the difference between the gross offtake and the net offtake
    shared_energy = df[GROSS_OFFTAKE] - net_offtake

    # Step 3: Assign the Rests back to the original injectors

    # The energy that is not shared after assignment
    # should be divided back to the original injectors
    # A ratio of the original injection should be used

    re_distributed_not_shared = (
        (df[GROSS_INJECTION].T / df[GROSS_INJECTION].sum(axis=1)) * not_shared_after_assignment
    ).T
    re_distributed_not_shared.fillna(0, inplace=True)

    # The nett injection is the sum of:
    # the injection that cannot be shared to begin with
    # (because participants cannot share with themselves)
    # and the injection that cannot be shared after assignment
    # (because participants cannot receive more than their offtake)

    net_injection = injection_that_cannot_be_shared + re_distributed_not_shared

    result = create_multi_index_output_frame(
        net_injection=net_injection, net_offtake=net_offtake, shared_energy=shared_energy
    )

    return result


def calculate(df: pd.DataFrame, method: CalculationMethod) -> pd.DataFrame:
    """Calculate the energy sharing for the given input data.

    This function is iterative if the method is optimal."""
    result = _calculate(df, method)

    if method in [CalculationMethod.FIXED, CalculationMethod.RELATIVE]:
        return result

    # Optimal method, we iterate until the amount of shared energy is 0
    final_result = result.copy()
    while not result[SHARED_ENERGY].eq(0).all().all():
        df = result_to_input_for_reiteration(result, df[KEY])
        result = _calculate(df, method)

        # Add the result to the final result
        # Overwrite NET_INJECTION and NET_OFFTAKE, Sum SHARED_ENERGY
        final_result[NET_INJECTION] = result[NET_INJECTION]
        final_result[NET_OFFTAKE] = result[NET_OFFTAKE]
        final_result[SHARED_ENERGY] += result[SHARED_ENERGY]

    return final_result



"""Data models for the Open Energy ID."""

import datetime as dt
from typing import overload

from typing import Self

import pandas as pd
from pydantic import BaseModel
import polars as pl


class TimeSeriesBase(BaseModel):
    """Pydantic base model for time series data."""

    index: list[dt.datetime]

    @classmethod
    def from_pandas(cls, data: pd.Series | pd.DataFrame) -> Self:
        """Create from a Pandas Object."""
        raise NotImplementedError

    def to_pandas(self, timezone: str = "UTC") -> pd.Series | pd.DataFrame:
        """Convert to a Pandas Object."""
        raise NotImplementedError

    @overload
    def to_json(self, path: None = None, **kwargs) -> str:
        """Dump to a JSON string."""

    @overload
    def to_json(self, path: str, **kwargs) -> None:
        """Dump to a JSON file."""

    def to_json(self, path: str | None = None, **kwargs) -> str | None:
        """Dump to a JSON string or file."""
        if path is None:
            return self.model_dump_json(**kwargs)
        encoding = kwargs.pop("encoding", "UTF-8")
        with open(path, "w", encoding=encoding) as file:
            file.write(self.model_dump_json(**kwargs))
        return None

    @overload
    @classmethod
    def from_json(cls, string: str, **kwargs) -> Self:
        """Load from a JSON string."""

    @overload
    @classmethod
    def from_json(cls, *, path: str, **kwargs) -> Self:
        """Load from a JSON file."""

    @classmethod
    def from_json(cls, string: str | None = None, path: str | None = None, **kwargs) -> Self:
        """Load from a JSON file or string."""
        if string:
            return cls.model_validate_json(string, **kwargs)
        if path:
            encoding = kwargs.pop("encoding", "UTF-8")
            with open(path, encoding=encoding) as file:
                return cls.model_validate_json(file.read(), **kwargs)
        raise ValueError("Either string or path must be provided.")


class TimeSeries(TimeSeriesBase):
    """
    Represents a time series data.
    Attributes:
        name (str | None): The name of the time series.
        data (list[float | None]): The data points of the time series.
    Methods:
        replace_nan_with_none(cls, data: list[float]) -> list[float | None]:
            Replace NaN values with None.
        from_pandas(cls, data: pd.Series) -> Self:
            Create a TimeSeries object from a Pandas Series.
        to_pandas(self, timezone: str = "UTC") -> pd.Series:
            Convert the TimeSeries object to a Pandas Series.
        from_polars(cls, data: pl.DataFrame | pl.LazyFrame) -> Self:
            Create a TimeSeries object from Polars data.
        to_polars(self, timezone: str = "UTC") -> pl.LazyFrame:
            Convert the TimeSeries object to a Polars LazyFrame.
    """

    name: str | None = None
    data: list[float | None]

    @classmethod
    def from_pandas(cls, data: pd.Series) -> Self:
        """Create from a Pandas Series."""
        if not hasattr(data, "name") or data.name is None:
            name = None
        else:
            name = str(data.name)
        return cls(name=name, data=data.tolist(), index=data.index.tolist())

    def to_pandas(self, timezone: str = "UTC") -> pd.Series:
        """Convert to a Pandas Series."""
        series = pd.Series(self.data, name=self.name, index=self.index)
        series.index = pd.to_datetime(series.index, utc=True)
        return series.tz_convert(timezone)

    @classmethod
    def from_polars(cls, data: pl.DataFrame | pl.LazyFrame) -> Self:
        """Create from Polars data."""
        # Always work with DataFrame
        df = data.collect() if isinstance(data, pl.LazyFrame) else data

        if len(df.columns) != 2:
            raise ValueError("Must contain exactly two columns: timestamp and value")

        value_col = [col for col in df.columns if col != "timestamp"][0]
        return cls(
            name=value_col,
            data=df[value_col].cast(pl.Float64).to_list(),  # Ensure float type
            index=df["timestamp"].cast(pl.Datetime).dt.convert_time_zone("UTC").to_list(),
        )

    def to_polars(self, timezone: str = "UTC") -> pl.LazyFrame:
        """Convert to Polars LazyFrame."""
        # Always return LazyFrame as specified in return type
        df = pl.DataFrame(
            {
                "timestamp": pl.Series(self.index, dtype=pl.Datetime).dt.convert_time_zone(
                    timezone
                ),
                "total" if self.name is None else self.name: pl.Series(self.data, dtype=pl.Float64),
            }
        )
        return df.lazy()


class TimeDataFrame(TimeSeriesBase):
    """Time series data with multiple columns."""

    columns: list[str]
    data: list[list[float | None]]

    @classmethod
    def from_pandas(cls, data: pd.DataFrame) -> Self:
        """Create from a Pandas DataFrame."""
        # Cast values to float | None
        values = [
            [float(x) if pd.notnull(x) else None for x in row] for row in data.values.tolist()
        ]
        return cls(columns=data.columns.tolist(), data=values, index=data.index.tolist())

    def to_pandas(self, timezone: str = "UTC") -> pd.DataFrame:
        """Convert to a Pandas DataFrame."""
        frame = pd.DataFrame(self.data, columns=self.columns, index=self.index)
        frame.index = pd.to_datetime(frame.index, utc=True)
        return frame.tz_convert(timezone)

    @classmethod
    def from_timeseries(cls, data: list[TimeSeries]) -> Self:
        """Create from a list of TimeSeries objects."""
        return cls(
            columns=[series.name or "" for series in data],  # Handle None names
            data=[series.data for series in data],  # Pass list of value lists
            index=data[0].index,
        )

    def to_timeseries(self) -> list[TimeSeries]:
        """Convert to a list of TimeSeries objects."""
        return [
            TimeSeries(name=col, data=[row[i] for row in self.data], index=self.index)
            for i, col in enumerate(self.columns)
        ]



"""Power Offtake peak analysis module."""

from .models import CapacityInput, CapacityOutput, PeakDetail
from .main import CapacityAnalysis

__all__ = ["CapacityInput", "CapacityAnalysis", "CapacityOutput", "PeakDetail"]



"""Constants for the Open Energy ID package."""

from typing import Literal

# METRICS

ELECTRICITY_DELIVERED: Literal["electricity_delivered"] = "electricity_delivered"
ELECTRICITY_EXPORTED: Literal["electricity_exported"] = "electricity_exported"
ELECTRICITY_PRODUCED: Literal["electricity_produced"] = "electricity_produced"

PRICE_DAY_AHEAD: Literal["price_day_ahead"] = "price_day_ahead"
PRICE_IMBALANCE_UPWARD: Literal["price_imbalance_upward"] = "price_imbalance_upward"
PRICE_IMBALANCE_DOWNWARD: Literal["price_imbalance_downward"] = "price_imbalance_downward"
PRICE_ELECTRICITY_DELIVERED: Literal["price_electricity_delivered"] = "price_electricity_delivered"
PRICE_ELECTRICITY_EXPORTED: Literal["price_electricity_exported"] = "price_electricity_exported"

RLP: Literal["RLP"] = "RLP"
SPP: Literal["SPP"] = "SPP"



"""Energy Sharing package."""

from .main import calculate
from .models import CalculationMethod, EnergySharingInput, EnergySharingOutput, KeyInput

__all__ = [
    "calculate",
    "CalculationMethod",
    "EnergySharingInput",
    "EnergySharingOutput",
    "KeyInput",
]



"""Multi-variable linear regression based on statsmodels
and Ordinary Least Squares (ols)."""

import numpy as np
import pandas as pd
import statsmodels.formula.api as fm
from patsy import LookupFactor, ModelDesc, Term  # pylint: disable=no-name-in-module
from statsmodels.sandbox.regression.predstd import wls_prediction_std

from openenergyid.enums import Granularity


class MultiVariableLinearRegression:
    """Multi-variable linear regression.

    Based on statsmodels and Ordinary Least Squares (ols).

    Pass a dataframe with the variable to be modelled y (dependent variable)
    and the possible independent variables x.
    Specify as string the name of the dependent variable, and optionally pass a list with names of
    independent variables to try
    (by default all other columns will be tried as independent variables).

    The analysis is based on a forward-selection approach: starting from a simple model,
    the model is iteratively refined and verified until no statistical relevant improvements
    can be obtained.
    Each model in the iteration loop is stored in the attribute self.list_of_fits.
    The selected model is self.fit (=pointer to the last element of self.list_of_fits).

    The dataframe can contain daily, weekly, monthly, yearly ... values.  Each row is an instance.
    """

    def __init__(
        self,
        data: pd.DataFrame,
        y: str,
        p_max: float = 0.05,
        list_of_x: list[str] = None,
        confint: float = 0.95,
        cross_validation: bool = False,
        allow_negative_predictions: bool = False,
        granularity: Granularity = None,
        single_use_exog_prefixes: list[str] = None,
        exogs__disallow_negative_coefficient: list[str] = None,
    ):
        """Parameters
        ----------
        data : TimeSeries
            Datetimeindex and both independent variables (x) and dependent variable (y) as columns
        y : str
            Name of the dependent (endogeneous) variable to model
        p_max : float (default=0.05)
            Acceptable p-value of the t-statistic for estimated parameters
        list_of_x : list of str (default=None)
            If None (default), try to build a model with all columns in the dataframe
            If a list with column names is given, only try these columns as independent variables
        confint : float, default=0.95
            Two-sided confidence interval for predictions.
        cross_validation : bool, default=False
            If True, compute the model based on cross-validation (leave one out)
            Only possible if the df has less than 15 entries.
            Note: this will take much longer computation times!
        allow_negative_predictions : bool, default=False
            If True, allow predictions to be negative.
            For gas consumption or PV production, this is not physical
            so allow_negative_predictions should be False
        granularity : Granularity, default=None
            Granularity of the data. Is only used for the output of the model.
            If None, the granularity is not set.
        single_use_exog_prefixes : list of str, default=None
            List of variable prefixes that indicate a variable type that should only be used once.
            For example, if the list contains "HDD", only one of the columns "HDD1", "HDD2", "HDD3" etc.
            will be used as an independent variable.
            Once the best fit using a variable with a given prefix is found, the other variables with the same
            prefix will not be used as independent variables.
        exogs__disallow_negative_coefficient : list of str, default=None
            List of variable names for which the coefficient is not allowed to be negative.
        """
        self.data = data.copy()
        if y not in self.data.columns:
            raise AssertionError(
                f"The dependent variable {y} is not a column in the dataframe",
            )
        self.y = y

        self.p_max = p_max
        self.list_of_x = list_of_x or [x for x in self.data.columns if x != self.y]
        self.confint = confint
        self.cross_validation = cross_validation
        self.allow_negative_predictions = allow_negative_predictions
        self.granularity = granularity
        self.single_use_exog_prefixes = single_use_exog_prefixes
        self.exogs__disallow_negative_coefficient = exogs__disallow_negative_coefficient
        self._fit = None
        self._list_of_fits = []
        self.list_of_cverrors = []

    @property
    def fit(self) -> fm.ols:
        """Fits a model to the data.

        Returns
        -------
            The fitted model.

        Raises
        ------
            UnboundLocalError: If `do_analysis()` has not been run before calling `fit()`.
        """
        if self._fit is None:
            raise UnboundLocalError(
                'Run "do_analysis()" first to fit a model to the data.',
            )
        else:
            return self._fit

    @property
    def list_of_fits(self) -> list[fm.ols]:
        """Returns the list of fits generated by the model.

        Raises
        ------
            UnboundLocalError: If the model has not been fitted yet.

        Returns
        -------
            list: The list of fits generated by the model.
        """
        if not self._list_of_fits:
            raise UnboundLocalError(
                'Run "do_analysis()" first to fit a model to the data.',
            )
        else:
            return self._list_of_fits

    def do_analysis(self):
        """Find the best model (fit) and create self.list_of_fits and self.fit"""
        if self.cross_validation:
            return self._do_analysis_cross_validation()
        else:
            return self._do_analysis_no_cross_validation()

    def _do_analysis_no_cross_validation(self):
        """Find the best model (fit) and create self.list_of_fits and self.fit"""
        # first model is just the mean
        response_term = [Term([LookupFactor(self.y)])]
        model_terms = [Term([])]  # empty term is the intercept
        all_model_terms_dict = {x: Term([LookupFactor(x)]) for x in self.list_of_x}
        # ...then add another term for each candidate
        # model_terms += [Term([LookupFactor(c)]) for c in candidates]
        model_desc = ModelDesc(response_term, model_terms)
        self._list_of_fits.append(fm.ols(model_desc, data=self.data).fit())
        # try to improve the model until no improvements can be found

        while all_model_terms_dict:
            # try each x and overwrite the best_fit if we find a better one
            # the first best_fit is the one from the previous round
            ref_fit = self._list_of_fits[-1]
            best_fit = self._list_of_fits[-1]
            best_bic = best_fit.bic
            for x, term in all_model_terms_dict.items():
                # make new_fit, compare with best found so far
                model_desc = ModelDesc(
                    response_term,
                    ref_fit.model.formula.rhs_termlist + [term],
                )
                fit = fm.ols(model_desc, data=self.data).fit()

                # Check if the coefficient of the variable is allowed to be negative
                if (
                    self.exogs__disallow_negative_coefficient is not None
                    and x in self.exogs__disallow_negative_coefficient
                    and fit.params[x] < 0
                ):
                    continue

                if fit.bic < best_bic:
                    best_bic = fit.bic
                    best_fit = fit
                    best_x = x
            # Sometimes, the obtained fit may be better, but contains unsignificant parameters.
            # Correct the fit by removing the unsignificant parameters and estimate again
            best_fit = self._prune(best_fit, p_max=self.p_max)

            # if best_fit does not contain more variables than ref fit, exit
            if len(best_fit.model.formula.rhs_termlist) == len(
                ref_fit.model.formula.rhs_termlist,
            ):
                break

            self._list_of_fits.append(best_fit)
            all_model_terms_dict.pop(best_x)

            # Check if `best_x` starts with a prefix that should only be used once
            # If so, remove all other variables with the same prefix from the list of candidates
            if self.single_use_exog_prefixes:
                for prefix in self.single_use_exog_prefixes:
                    if best_x.startswith(prefix):
                        all_model_terms_dict = {
                            k: v
                            for k, v in all_model_terms_dict.items()
                            if not k.startswith(prefix)
                        }

        self._fit = self._list_of_fits[-1]

    def _do_analysis_cross_validation(self):
        """Find the best model (fit) based on cross-valiation (leave one out)"""
        assert len(self.data) < 15, (
            "Cross-validation is not implemented if your sample contains more than 15 datapoints"
        )

        # initialization: first model is the mean, but compute cv correctly.
        errors = []
        response_term = [Term([LookupFactor(self.y)])]
        model_terms = [Term([])]  # empty term is the intercept
        model_desc = ModelDesc(response_term, model_terms)
        for i in self.data.index:
            # make new_fit, compute cross-validation and store error
            df_ = self.data.drop(i, axis=0)
            fit = fm.ols(model_desc, data=df_).fit()
            cross_prediction = self._predict(fit=fit, data=self.data.loc[[i], :])
            errors.append(cross_prediction["predicted"] - cross_prediction[self.y])

        self._list_of_fits = [fm.ols(model_desc, data=self.data).fit()]
        self.list_of_cverrors = [np.mean(np.abs(np.array(errors)))]

        # try to improve the model until no improvements can be found
        all_model_terms_dict = {x: Term([LookupFactor(x)]) for x in self.list_of_x}
        while all_model_terms_dict:
            # import pdb;pdb.set_trace()
            # try each x in all_exog and overwrite if we find a better one
            # at the end of iteration (and not earlier), save the best of the iteration
            better_model_found = False
            best = dict(fit=self._list_of_fits[-1], cverror=self.list_of_cverrors[-1])
            for x, term in all_model_terms_dict.items():
                model_desc = ModelDesc(
                    response_term,
                    self._list_of_fits[-1].model.formula.rhs_termlist + [term],
                )
                # cross_validation, currently only implemented for monthly data
                # compute the mean error for a given formula based on leave-one-out.
                errors = []
                for i in self.data.index:
                    # make new_fit, compute cross-validation and store error
                    df_ = self.data.drop(i, axis=0)
                    fit = fm.ols(model_desc, data=df_).fit()
                    cross_prediction = self._predict(
                        fit=fit,
                        data=self.data.loc[[i], :],
                    )
                    errors.append(
                        cross_prediction["predicted"] - cross_prediction[self.y],
                    )
                cverror = np.mean(np.abs(np.array(errors)))
                # compare the model with the current fit
                if cverror < best["cverror"]:
                    # better model, keep it
                    # first, reidentify using all the datapoints
                    best["fit"] = fm.ols(model_desc, data=self.data).fit()
                    best["cverror"] = cverror
                    better_model_found = True
                    best_x = x

            if better_model_found:
                self._list_of_fits.append(best["fit"])
                self.list_of_cverrors.append(best["cverror"])

            else:
                # if we did not find a better model, exit
                break

            # next iteration with the found exog removed
            all_model_terms_dict.pop(best_x)

            # Check if `best_x` starts with a prefix that should only be used once
            # If so, remove all other variables with the same prefix from the list of candidates
            if self.single_use_exog_prefixes:
                for prefix in self.single_use_exog_prefixes:
                    if best_x.startswith(prefix):
                        all_model_terms_dict = {
                            k: v
                            for k, v in all_model_terms_dict.items()
                            if not k.startswith(prefix)
                        }

        self._fit = self._list_of_fits[-1]

    def _prune(self, fit: fm.ols, p_max: float) -> fm.ols:
        """If the fit contains statistically insignificant parameters, remove them.
        Returns a pruned fit where all parameters have p-values of the t-statistic below p_max

        Parameters
        ----------
        fit: fm.ols fit object
            Can contain insignificant parameters
        p_max : float
            Maximum allowed probability of the t-statistic

        Returns
        -------
        fit: fm.ols fit object
            Won't contain any insignificant parameters

        """

        def remove_from_model_desc(x: str, model_desc: ModelDesc) -> ModelDesc:
            """Return a model_desc without x"""
            rhs_termlist = []
            for t in model_desc.rhs_termlist:
                if not t.factors:
                    # intercept, add anyway
                    rhs_termlist.append(t)
                elif x != t.factors[0]._varname:  # pylint: disable=protected-access
                    # this is not the term with x
                    rhs_termlist.append(t)

            md = ModelDesc(model_desc.lhs_termlist, rhs_termlist)
            return md

        corrected_model_desc = ModelDesc(
            fit.model.formula.lhs_termlist[:],
            fit.model.formula.rhs_termlist[:],
        )
        pars_to_prune = fit.pvalues.where(fit.pvalues > p_max).dropna().index.tolist()
        try:
            pars_to_prune.remove("Intercept")
        except ValueError:
            pass
        while pars_to_prune:
            corrected_model_desc = remove_from_model_desc(
                pars_to_prune[0],
                corrected_model_desc,
            )
            fit = fm.ols(corrected_model_desc, data=self.data).fit()
            pars_to_prune = fit.pvalues.where(fit.pvalues > p_max).dropna().index.tolist()
            try:
                pars_to_prune.remove("Intercept")
            except ValueError:
                pass
        return fit

    @staticmethod
    def find_best_rsquared(list_of_fits: list[fm.ols]) -> fm.ols:
        """Return the best fit, based on rsquared"""
        res = sorted(list_of_fits, key=lambda x: x.rsquared)
        return res[-1]

    @staticmethod
    def find_best_akaike(list_of_fits: list[fm.ols]) -> fm.ols:
        """Return the best fit, based on Akaike information criterion"""
        res = sorted(list_of_fits, key=lambda x: x.aic)
        return res[0]

    @staticmethod
    def find_best_bic(list_of_fits: list[fm.ols]) -> fm.ols:
        """Return the best fit, based on Akaike information criterion"""
        res = sorted(list_of_fits, key=lambda x: x.bic)
        return res[0]

    def _predict(self, fit: fm.ols, data: pd.DataFrame) -> pd.DataFrame:
        """Return a df with predictions and confidence interval

        Notes
        -----
        The df will contain the following columns:
        - 'predicted': the model output
        - 'interval_u', 'interval_l': upper and lower confidence bounds.
        The result will depend on the following attributes of self:
        confint : float (default=0.95)
            Confidence level for two-sided hypothesis
        allow_negative_predictions : bool (default=True)
            If False, correct negative predictions to zero
            (typically for energy consumption predictions)

        Parameters
        ----------
        fit : Statsmodels fit
        data : pandas DataFrame or None (default)
            If None, use self.data

        Returns
        -------
        result : pandas DataFrame
            Copy of df with additional columns 'predicted', 'interval_u' and 'interval_l'
        """
        # Add model results to data as column 'predictions'
        result = data.copy()
        if "Intercept" in fit.model.exog_names:
            result["Intercept"] = 1.0
        result["predicted"] = fit.predict(result)
        if not self.allow_negative_predictions:
            result.loc[result["predicted"] < 0, "predicted"] = 0

        _prstd, interval_l, interval_u = wls_prediction_std(
            fit,
            result[fit.model.exog_names],
            alpha=1 - self.confint,
        )
        result["interval_l"] = interval_l
        result["interval_u"] = interval_u

        if "Intercept" in result:
            result.drop(labels=["Intercept"], axis=1, inplace=True)

        return result

    def add_prediction(self):
        """Add predictions and confidence interval to self.df
        self.df will contain the following columns:
        - 'predicted': the model output
        - 'interval_u', 'interval_l': upper and lower confidence bounds.

        Parameters
        ----------
        None, but the result depends on the following attributes of self:
        confint : float (default=0.95)
            Confidence level for two-sided hypothesis
        allow_negative_predictions : bool (default=True)
            If False, correct negative predictions to zero
            (typically for energy consumption predictions)

        Returns
        -------
        Nothing, adds columns to self.df
        """
        self.data = self._predict(fit=self.fit, data=self.data)

    def validate(
        self, min_rsquared: float = 0.75, max_f_pvalue: float = 0.05, max_pvalues: float = 0.05
    ) -> bool:
        """Checks if the model is valid.

        Returns
        -------
            bool: True if the model is valid, False otherwise.
        """
        if self.fit.rsquared_adj < min_rsquared:
            return False

        if self.fit.f_pvalue > max_f_pvalue:
            return False

        param_keys = self.fit.pvalues.keys().tolist()
        param_keys.remove("Intercept")
        for k in param_keys:
            if self.fit.pvalues[k] > max_pvalues:
                return False

        return True



"""Main module of the DynTar package."""

from typing import cast
import pandas as pd

from openenergyid.const import (
    ELECTRICITY_DELIVERED,
    ELECTRICITY_EXPORTED,
    PRICE_ELECTRICITY_DELIVERED,
    PRICE_ELECTRICITY_EXPORTED,
    RLP,
    SPP,
)

from .const import (
    ELECTRICITY_DELIVERED_SMR3,
    ELECTRICITY_EXPORTED_SMR3,
    ELECTRICITY_DELIVERED_SMR2,
    ELECTRICITY_EXPORTED_SMR2,
    COST_ELECTRICITY_DELIVERED_SMR2,
    COST_ELECTRICITY_EXPORTED_SMR2,
    COST_ELECTRICITY_DELIVERED_SMR3,
    COST_ELECTRICITY_EXPORTED_SMR3,
    RLP_WEIGHTED_PRICE_DELIVERED,
    SPP_WEIGHTED_PRICE_EXPORTED,
    HEATMAP_DELIVERED,
    HEATMAP_EXPORTED,
    HEATMAP_TOTAL,
    HEATMAP_DELIVERED_DESCRIPTION,
    HEATMAP_EXPORTED_DESCRIPTION,
    HEATMAP_TOTAL_DESCRIPTION,
    Register,
)


def weigh_by_monthly_profile(df: pd.DataFrame, series_name, profile_name) -> pd.Series:
    """Weigh a time series by a monthly profile."""
    grouped = df.groupby(pd.Grouper(freq="MS"))
    return grouped[series_name].transform("sum") * grouped[profile_name].transform(
        lambda x: x / x.sum()
    )


def extend_dataframe_with_smr2(
    df: pd.DataFrame,
    inplace: bool = False,
    registers: list[Register] | None = None,
) -> pd.DataFrame | None:
    """Extend a DataFrame with the SMR2 columns."""
    if not inplace:
        result_df = df.copy()
    else:
        result_df = df

    if registers is None:
        registers = [Register.DELIVERY, Register.EXPORT]

    if Register.DELIVERY in registers:
        result_df[ELECTRICITY_DELIVERED_SMR2] = weigh_by_monthly_profile(
            df, ELECTRICITY_DELIVERED, RLP
        )
    if Register.EXPORT in registers:
        result_df[ELECTRICITY_EXPORTED_SMR2] = weigh_by_monthly_profile(
            df, ELECTRICITY_EXPORTED, SPP
        )

    result_df.rename(
        columns={
            ELECTRICITY_DELIVERED: ELECTRICITY_DELIVERED_SMR3,
            ELECTRICITY_EXPORTED: ELECTRICITY_EXPORTED_SMR3,
        },
        inplace=True,
        errors="ignore",
    )

    if not inplace:
        return result_df
    return None


def extend_dataframe_with_costs(
    df: pd.DataFrame, inplace: bool = False, registers: list[Register] | None = None
) -> pd.DataFrame | None:
    """Extend a DataFrame with the cost columns."""
    if not inplace:
        result_df = df.copy()
    else:
        result_df = df

    if registers is None:
        registers = [Register.DELIVERY, Register.EXPORT]

    if Register.DELIVERY in registers:
        result_df[COST_ELECTRICITY_DELIVERED_SMR2] = (
            df[ELECTRICITY_DELIVERED_SMR2] * df[PRICE_ELECTRICITY_DELIVERED]
        )
        result_df[COST_ELECTRICITY_DELIVERED_SMR3] = (
            df[ELECTRICITY_DELIVERED_SMR3] * df[PRICE_ELECTRICITY_DELIVERED]
        )

    if Register.EXPORT in registers:
        result_df[COST_ELECTRICITY_EXPORTED_SMR2] = (
            df[ELECTRICITY_EXPORTED_SMR2] * df[PRICE_ELECTRICITY_EXPORTED] * -1
        )
        result_df[COST_ELECTRICITY_EXPORTED_SMR3] = (
            df[ELECTRICITY_EXPORTED_SMR3] * df[PRICE_ELECTRICITY_EXPORTED] * -1
        )

    if not inplace:
        return result_df
    return None


def extend_dataframe_with_weighted_prices(
    df: pd.DataFrame, inplace: bool = False, registers: list[Register] | None = None
) -> pd.DataFrame | None:
    """Extend a DataFrame with the weighted price columns."""
    if not inplace:
        df = df.copy()

    if registers is None:
        registers = [Register.DELIVERY, Register.EXPORT]

    if Register.DELIVERY in registers:
        rlp_weighted_price_delivered = (df[PRICE_ELECTRICITY_DELIVERED] * df[RLP]).resample(
            "MS"
        ).sum() / df[RLP].resample("MS").sum()
        df[RLP_WEIGHTED_PRICE_DELIVERED] = rlp_weighted_price_delivered.reindex_like(
            df[RLP], method="ffill"
        )

    if Register.EXPORT in registers:
        spp_weighted_price_exported = (df[PRICE_ELECTRICITY_EXPORTED] * df[SPP]).resample(
            "MS"
        ).sum() / df[SPP].resample("MS").sum()
        df[SPP_WEIGHTED_PRICE_EXPORTED] = spp_weighted_price_exported.reindex_like(
            df[SPP], method="ffill"
        )

    if not inplace:
        return df
    return None


def extend_dataframe_with_heatmap(
    df: pd.DataFrame, inplace: bool = False, registers: list[Register] | None = None
) -> pd.DataFrame | None:
    """Extend a DataFrame with the heatmap columns."""
    if not inplace:
        df = df.copy()

    if registers is None:
        registers = [Register.DELIVERY, Register.EXPORT]

    if Register.DELIVERY in registers:
        energy_delta_delivered = df[ELECTRICITY_DELIVERED_SMR2] - df[ELECTRICITY_DELIVERED_SMR3]
        price_delta_delivered = df[RLP_WEIGHTED_PRICE_DELIVERED] - df[PRICE_ELECTRICITY_DELIVERED]
        heatmap_score_delivered = energy_delta_delivered * price_delta_delivered
        heatmap_score_delivered.fillna(0, inplace=True)
        # Invert score so that positive values indicate a positive impact
        heatmap_score_delivered = -heatmap_score_delivered
        df[HEATMAP_DELIVERED] = heatmap_score_delivered

    if Register.EXPORT in registers:
        energy_delta_exported = df[ELECTRICITY_EXPORTED_SMR2] - df[ELECTRICITY_EXPORTED_SMR3]
        price_delta_exported = df[SPP_WEIGHTED_PRICE_EXPORTED] - df[PRICE_ELECTRICITY_EXPORTED]
        heatmap_score_exported = energy_delta_exported * price_delta_exported
        heatmap_score_exported.fillna(0, inplace=True)
        df[HEATMAP_EXPORTED] = heatmap_score_exported

    if Register.DELIVERY in registers and Register.EXPORT in registers:
        heatmap_score_delivered = cast(pd.Series, df[HEATMAP_DELIVERED])
        heatmap_score_exported = cast(pd.Series, df[HEATMAP_EXPORTED])
        heatmap_score_combined = heatmap_score_delivered + heatmap_score_exported
    elif Register.DELIVERY in registers:
        heatmap_score_combined = heatmap_score_delivered
    else:
        heatmap_score_combined = heatmap_score_exported
    df[HEATMAP_TOTAL] = heatmap_score_combined

    if not inplace:
        return df
    return None


def map_delivery_description(
    price_delivered, price_rlp, electricity_delivered_smr3, electricity_delivered_smr2
):
    """Map the delivery description."""
    if price_delivered > price_rlp and electricity_delivered_smr3 > electricity_delivered_smr2:
        return 1
    if price_delivered > price_rlp and electricity_delivered_smr3 < electricity_delivered_smr2:
        return 2
    if price_delivered < price_rlp and electricity_delivered_smr3 > electricity_delivered_smr2:
        return 3
    if price_delivered < price_rlp and electricity_delivered_smr3 < electricity_delivered_smr2:
        return 4
    return 0


def map_export_description(
    price_exported, price_spp, electricity_exported_smr3, electricity_exported_smr2
):
    """Map the export description."""
    if price_exported > price_spp and electricity_exported_smr3 > electricity_exported_smr2:
        return 5
    if price_exported > price_spp and electricity_exported_smr3 < electricity_exported_smr2:
        return 6
    if price_exported < price_spp and electricity_exported_smr3 > electricity_exported_smr2:
        return 7
    if price_exported < price_spp and electricity_exported_smr3 < electricity_exported_smr2:
        return 8
    return 0


def map_total_description(
    abs_heatmap_delivered, abs_heatmap_exported, delivered_description, exported_description
):
    """Map the total description."""
    if abs_heatmap_delivered > abs_heatmap_exported:
        return delivered_description
    return exported_description


def extend_dataframe_with_heatmap_description(
    df: pd.DataFrame, inplace: bool = False, registers: list[Register] | None = None
) -> pd.DataFrame | None:
    """Extend a DataFrame with the heatmap description columns."""
    if not inplace:
        df = df.copy()

    if registers is None:
        registers = [Register.DELIVERY, Register.EXPORT]

    if Register.DELIVERY in registers:
        df[HEATMAP_DELIVERED_DESCRIPTION] = list(
            map(
                map_delivery_description,
                df[PRICE_ELECTRICITY_DELIVERED],
                df[RLP_WEIGHTED_PRICE_DELIVERED],
                df[ELECTRICITY_DELIVERED_SMR3],
                df[ELECTRICITY_DELIVERED_SMR2],
            )
        )

    if Register.EXPORT in registers:
        df[HEATMAP_EXPORTED_DESCRIPTION] = list(
            map(
                map_export_description,
                df[PRICE_ELECTRICITY_EXPORTED],
                df[SPP_WEIGHTED_PRICE_EXPORTED],
                df[ELECTRICITY_EXPORTED_SMR3],
                df[ELECTRICITY_EXPORTED_SMR2],
            )
        )

    if Register.DELIVERY in registers and Register.EXPORT in registers:
        df[HEATMAP_TOTAL_DESCRIPTION] = list(
            map(
                map_total_description,
                df[HEATMAP_DELIVERED].abs(),
                df[HEATMAP_EXPORTED].abs(),
                df[HEATMAP_DELIVERED_DESCRIPTION],
                df[HEATMAP_EXPORTED_DESCRIPTION],
            )
        )
    elif Register.DELIVERY in registers:
        df[HEATMAP_TOTAL_DESCRIPTION] = df[HEATMAP_DELIVERED_DESCRIPTION]
    else:
        df[HEATMAP_TOTAL_DESCRIPTION] = df[HEATMAP_EXPORTED_DESCRIPTION]

    if not inplace:
        return df


def calculate_dyntar_columns(
    df: pd.DataFrame,
    inplace: bool = False,
    registers: list[Register] | None = None,
) -> pd.DataFrame | None:
    """Calculate all columns required for the dynamic tariff analysis."""
    if not inplace:
        df = df.copy()

    if registers is None:
        registers = [Register.DELIVERY, Register.EXPORT]

    extend_dataframe_with_smr2(df, inplace=True, registers=registers)
    extend_dataframe_with_costs(df, inplace=True, registers=registers)
    extend_dataframe_with_weighted_prices(df, inplace=True, registers=registers)
    extend_dataframe_with_heatmap(df, inplace=True, registers=registers)
    extend_dataframe_with_heatmap_description(df, inplace=True, registers=registers)

    if not inplace:
        return df
    return None


def summarize_result(df: pd.DataFrame) -> pd.Series:
    """Summarize the dynamic tariff analysis result."""
    summary = df.filter(like="cost").sum()

    abs_smr2 = summary.filter(like="smr2").abs().sum()

    summary["cost_electricity_total_smr2"] = summary.filter(like="smr2").sum()
    summary["cost_electricity_total_smr3"] = summary.filter(like="smr3").sum()

    summary["ratio"] = (
        summary["cost_electricity_total_smr3"] - summary["cost_electricity_total_smr2"]
    ) / abs_smr2

    return summary
